import os
import sys
import re
import pwd
import csv
import ConfigParser
import tempfile
import operator
from stat import S_IRUSR, S_IWUSR, S_IRGRP, S_IROTH, ST_DEV

from clquota.utils import check_command, run_command


class NoSuchPackageException(Exception):
    def __init__(self, package):
        Exception.__init__(self, "No such package (%s)" % (package,))


class NoSuchUserException(Exception):
    def __init__(self, user):
        Exception.__init__(self, "No such user (%s)" % (user,))


class InsufficientPrivilegesException(Exception):
    def __init__(self):
        Exception.__init__(self, "Insufficient privileges")

class IncorrectLimitFormatException(Exception):
    def __init__(self, limit):
        Exception.__init__(self, "Incorrect limit format (%s)" %(limit,))

class GeneralException(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)

class QuotaDisabledException(Exception):
    def __init__(self, uid, homedir):
        Exception.__init__(self, "Quota disabled for %s (%s)" %(uid, homedir))


class QuotaWrapper(object):
    '''
    Base quota class for inode quotas handling
    '''
    MOUNT     = '/bin/mount'
    SETQUOTA  = '/usr/sbin/setquota'
    REPQUOTA  = '/usr/sbin/repquota'
    GETPACKS  = '/usr/bin/getcontrolpaneluserspackages'
    DATAFILE  = '/etc/container/cl-quotas.dat'
    CACHEFILE = '/etc/container/cl-quotas.cache'

    def __init__(self):
        check_command(QuotaWrapper.GETPACKS)
        check_command(QuotaWrapper.MOUNT)
        check_command(QuotaWrapper.REPQUOTA)
        check_command(QuotaWrapper.SETQUOTA)
        self._grace      = {}
        self._quota      = {}
        self._devices    = {}
        self._package_to_uids_map = {}
        self._uid_to_packages_map = {}
        self._uid_to_homedir_map  = {}
        self._dh         = self._get_saved_data_handler()
        self._fields     = ['bytes_used', 'bytes_soft',
            'bytes_hard', 'inodes_used', 'inodes_soft', 'inodes_hard']

    def get_user_limits(self, uid):
        '''
        Returns user limits converted to tuples
        '''
        return self._convert_data_to_tuples(self._get_current_quotas(uid))

    def get_all_users_limits(self):
        '''
        Returns all user limits converted to tuples
        '''
        return self._convert_data_to_tuples(self._get_current_quotas())

    def get_package_limits(self, package):
        '''
        Returns package limits converted to tuples
        '''
        return self._convert_data_to_tuples(self._get_package_quotas(package))

    def get_all_packages_limits(self):
        '''
        Returns all packages limits converted to tuples
        '''
        return self._convert_data_to_tuples(self._get_package_quotas())

    def set_user_limit(self, uid, soft=None, hard=None, save=True, sync=True):
        '''
        Sets limits for users
        '''
        self._check_admin()
        soft = self._check_limit(soft)
        hard = self._check_limit(hard)
        saved_quotas = self._get_current_quotas(uid)
        if uid == '0':
            self._apply_to_all_if_not_set(soft=soft, hard=hard, save=save)
            return
        user_soft, user_hard = soft, hard
        if sync:
            user_soft, user_hard = self._combine_user_limits(
                uid=uid, soft=soft, hard=hard)
        cmd = [
            QuotaWrapper.SETQUOTA,
            '-u', uid,
            saved_quotas[uid]['bytes_soft'],
            saved_quotas[uid]['bytes_hard'],
            user_soft or '0',
            user_hard or '0',
            self._get_home_device(self._fetch_homedir(uid))
        ]
        run_command(cmd)
        if save:
            self._save_user_limits(uid, soft, hard)

    def set_package_limit(self, package,
            soft=None, hard=None, save=True, sync=True):
        '''
        Sets limits for packages
        '''
        self._check_admin()
        soft = self._check_limit(soft)
        hard = self._check_limit(hard)
        std_in = []
        try:
            self._get_package_to_users_map(package)
        except NoSuchPackageException:
            if sync:
                return
        device_user_map = self._get_device_user_map()
        if sync:
            soft, hard = self._get_saved_package_limits_if_none(
                package, soft, hard)
        saved_quotas = self._get_current_quotas()
        for device in device_user_map.keys():
            cmd = [QuotaWrapper.SETQUOTA, '-bu', device]
            for uid in self._get_package_to_users_map(package):
                if uid not in device_user_map[device]:
                    continue
                data = self._combine_package_limits(
                    uid=uid, soft=soft, hard=hard)
                if not data:
                    continue
                std_in.append(
                    '%s %s %s %s %s' % (
                        uid,
                        saved_quotas[uid]['bytes_soft'],
                        saved_quotas[uid]['bytes_hard'],
                        data[0] or '0',
                        data[1] or '0'))
            if len(std_in) == 0:
                continue
            std_in = ('\n'.join(std_in) + '\n')
            run_command(cmd, std_in)
            std_in = []
        if save:
            self._save_package_limits(package, soft, hard)

    def synchronize(self):
        '''
        Read limits from file and applies them to packages and users
        '''
        self._check_admin()
        if self._dh.has_section('packages'):
            for package in self._dh.options('packages'):
                try:
                    self._get_package_to_users_map(package)
                    limits = self._dh.get('packages', package).split(':')
                    if len(limits) < 2:
                        continue
                    soft, hard = limits
                    self.set_package_limit(package, soft, hard, False)
                except NoSuchPackageException:
                    continue
            
        if self._dh.has_section('users'):
            for uid in self._dh.options('users'):
                try:
                    self._fetch_homedir(uid)
                    limits = self._dh.get('users', uid).split(':')
                    if len(limits) < 2:
                        continue
                    soft, hard = limits
                    self.set_user_limit(uid, soft, hard, False)
                except NoSuchUserException:
                    continue

    def save_user_cache(self):
        '''
        Caches the limits to non-privileged user to see them
        '''
        self._check_admin()
        cache_content = []
        current_quotas = self._get_current_quotas()
        for k in sorted(current_quotas.keys(), key=int):
            cache_content.append([k] + map((lambda x: current_quotas[k][x]),
                    self._fields))
        file_handler = self._prepare_writer(QuotaWrapper.CACHEFILE)
        csv_out = csv.writer(file_handler, quoting=csv.QUOTE_MINIMAL)
        csv_out.writerows(cache_content)
        self._end_writer(QuotaWrapper.CACHEFILE)

    def _check_admin(self):
        '''
        Raise exception if no admin user
        '''
        if os.geteuid() != 0:
            raise InsufficientPrivilegesException()

    def _get_saved_data_handler(self):
        '''
        Gets ConfigParser handler for future use
        '''
        dh = ConfigParser.ConfigParser()
        dh.optionxform = str
        dh.read(QuotaWrapper.DATAFILE)
        return dh

    def _get_device_user_map(self):
        '''
        Retuns dictionary mapping devices to lists of users
        '''
        devices_map = {}
        device_user_pairs = []
        for uid in self._get_list_of_uids():
            try:
                device_user_pairs.append(
                    (self._get_home_device(self._fetch_homedir(uid)), uid))
            except KeyError:
                continue
        for pair in device_user_pairs:
            if pair[0] not in devices_map:
                devices_map[pair[0]] = []
            devices_map[pair[0]].append(pair[1])
        return devices_map

    def _check_limit(self, limit):
        if limit is None:
            return limit
        limit_pattern = re.compile(r'(\d+)')
        pattern_match = limit_pattern.search(limit)
        if not pattern_match:
            raise IncorrectLimitFormatException(limit)
        return pattern_match.group(1)

    def _apply_to_all_if_not_set(self, soft=None, hard=None, save=True):
        '''
        Applies limits to all users if no other (user or package) ones has
        not been set for them
        '''
        std_in = []
        device_user_map = self._get_device_user_map()
        saved_quotas = self._get_current_quotas()
        for device in device_user_map.keys():
            cmd = [QuotaWrapper.SETQUOTA, '-bu', device]
            for uid in self._get_list_of_uids():
                if uid not in device_user_map[device]:
                    continue
                data = self._combine_default_limits(
                    uid=uid, soft=soft, hard=hard)
                if not data:
                    continue
                std_in.append('%s %s %s %s %s' % (
                    uid,
                    saved_quotas[uid]['bytes_soft'],
                    saved_quotas[uid]['bytes_hard'],
                    data[0] or '0',
                    data[1] or '0'))
            if len(std_in) == 0:
                continue
            std_in = ('\n'.join(std_in) + '\n')
            run_command(cmd, std_in)
            std_in = []
        if save:
            self._save_user_limits(uid='0', soft=soft, hard=hard)

    def _combine_user_limits(self, uid, soft=None, hard=None):
        '''
        Determines user limits taking into account
        saved package and default ones
        '''
        u_soft, u_hard = self._get_saved_user_limits_if_none(
            uid=uid, soft=soft, hard=hard)
        p_soft, p_hard = None, None
        for package in self._get_uid_to_packages_map(uid):
            p_soft, p_hard = self._get_saved_package_limits_if_none_or_unlim(
                    package=package, soft=u_soft, hard=u_hard)
        d_soft, d_hard = self._get_saved_user_limits_if_none_or_unlim(
                uid='0', soft=p_soft, hard=p_hard)
        return (d_soft, d_hard)

    def _combine_package_limits(self, uid, soft=None, hard=None):
        '''
        Determines package limits taking into account
        saved user and default ones
        '''
        t_soft, t_hard = None, None
        t_soft, t_hard = self._get_saved_user_limits_if_none(
            uid=uid, soft=t_soft, hard=t_hard)
        if t_soft and t_hard:
            return ()
        if soft is None:
            soft = t_soft
        if hard is None:
            hard = t_hard
        return self._get_saved_user_limits_if_none_or_unlim(
                uid='0', soft=soft, hard=hard)

    def _combine_default_limits(self, uid, soft=None, hard=None):
        '''
        Determines default limits taking into account
        saved user and package ones
        '''
        t_soft, t_hard = None, None
        t_soft, t_hard = self._get_saved_user_limits_if_none(
                uid=uid, soft=t_soft, hard=t_hard)
        for package in self._get_uid_to_packages_map(uid):
            t_soft, t_hard = self._get_saved_package_limits_if_none(
                package=package, soft=t_soft, hard=t_hard)
        if t_soft and t_hard:
            return ()
        if t_soft is None:
            t_soft = soft
        if t_hard is None:
            t_hard = hard
        return (soft, hard)

    def _get_saved_user_limits_if_none(self, uid, soft=None, hard=None):
        '''
        Applies saved user limits if none has passed
        '''
        try:
            user_soft, user_hard = self._dh.get('users', uid).split(':')
            if soft is None and user_soft != '0':
                soft = user_soft
            if hard is None and user_hard != '0':
                hard = user_hard
        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
            pass
        soft = self._check_limit(soft)
        hard = self._check_limit(hard)
        return (soft, hard)

    def _get_saved_user_limits_if_none_or_unlim(self, uid, soft=None, hard=None):
        '''
        Applies saved user limits if none or unlimit has been passed
        '''
        try:
            user_soft, user_hard = self._dh.get('users', uid).split(':')
            if (soft is None or soft == '0') and user_soft != '0':
                soft = user_soft
            if (hard is None or hard == '0') and user_hard != '0':
                hard = user_hard
        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
            pass
        soft = self._check_limit(soft)
        hard = self._check_limit(hard)
        return (soft, hard)

    def _get_saved_package_limits_if_none(self, package, soft=None, hard=None):
        '''
        Applies saved package limits if none has passed
        '''
        try:
            pack_soft, pack_hard = self._dh.get('packages', package).split(':')
            if soft is None and pack_soft != '0':
                soft = pack_soft
            if hard is None and pack_hard != '0':
                hard = pack_hard
        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
            pass
        soft = self._check_limit(soft)
        hard = self._check_limit(hard)
        return (soft, hard)

    def _get_saved_package_limits_if_none_or_unlim(
            self, package, soft=None, hard=None):
        '''
        Applies saved package limits if none or unlimit has been passed
        '''
        try:
            pack_soft, pack_hard = self._dh.get('packages', package).split(':')
            if (soft is None or soft == '0') and pack_soft != '0':
                soft = pack_soft
            if (hard is None or hard == '0') and pack_hard != '0':
                hard = pack_hard
        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
            pass
        soft = self._check_limit(soft)
        hard = self._check_limit(hard)
        return (soft, hard)

    def _get_current_quotas(self, uid=None):
        if os.geteuid() != 0:
            return self._load_user_cache()
        if not self._quota:
            self._quota = self._load_current_quotas()
        if uid:
            try:
                return {uid: self._quota[uid]}
            except KeyError:
                self._check_if_quota_enabled(uid)
                raise NoSuchUserException(uid)
        return self._quota

    def _get_package_quotas(self, packname=None):
        '''
        Prepares package limits data for outputting
        '''
        q = {}
        for package in self._get_list_of_packages():
            values = ['-']
            try:
                soft, hard = self._dh.get('packages', package).split(':')
                soft = self._check_limit(soft)
                hard = self._check_limit(hard)
                values.extend([soft, hard])
            except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
                values.extend(['0', '0'])
            q.update(self._populate(package, values))
        if packname:
            try:
                return {packname: q[packname]}
            except KeyError:
                raise NoSuchPackageException(packname)
        return q

    def _populate(self, item, data):
        return {item: dict(map((lambda x: (x[1], data[x[0]])),
            enumerate(self._fields[3:])))}

    def _get_list_of_packages(self):
        return self._get_package_to_users_map().keys()

    def _get_list_of_uids(self):
        return self._get_uid_to_packages_map().keys()

    def _get_package_to_users_map(self, package=None):
        if not self._package_to_uids_map:
            self._package_to_uids_map = self._load_package_uids_data()
        if package:
            try:
                return self._package_to_uids_map[package]
            except KeyError:
                raise NoSuchPackageException(package)
        return self._package_to_uids_map

    def _check_if_quota_enabled(self, uid):
        devices = self._get_quota_devices()
        home_dir = self._fetch_homedir(uid)
        if (len(devices) == 0
                or self._get_home_device(home_dir) not in devices):
            raise QuotaDisabledException(uid, home_dir)

    def _get_uid_to_packages_map(self, uid=None):
        if not self._uid_to_packages_map:
            self._package_to_uids_map = self._load_package_uids_data()
        if uid:
            try:
                return self._uid_to_packages_map[uid]
            except KeyError:
                
                raise NoSuchUserException(uid)
        return self._uid_to_packages_map

    def _load_package_uids_data(self):
        '''
        Gets map of packages and users
        '''
        packages = {}
        if os.geteuid() != 0:
            return packages
        cmd = [QuotaWrapper.GETPACKS, '--list-all']
        for entry in run_command(cmd).splitlines():
            uid, package = map((lambda x: x.strip()), entry.split(' ', 1))
            if not package in packages:
                packages[package] = []
            packages[package].append(uid)
            if not uid in self._uid_to_packages_map:
                self._uid_to_packages_map[uid] = []
            self._uid_to_packages_map[uid].append(package)
        cmd = [QuotaWrapper.GETPACKS, '--list-packages']
        for package in run_command(cmd).splitlines():
             if package not in packages:
                packages[package] = []
        if 'default' not in packages:
            packages['default'] = []
        return packages

    def _get_quota_devices(self):
        if not self._devices:
            self._devices = self._load_quota_devices()
        return self._devices

    def _convert_data_to_tuples(self, data):
        '''
        Convert dict to tuples for passing to printing routines
        '''
        for key in data.keys():
            try:
                entry = tuple(map((lambda x: (x, data[key][x])),
                    self._fields[3:]))
                data[key] = entry
            except KeyError:
                continue
        return data

    def _load_current_quotas(self):
        '''
        Gets current quota settings for further processing
        '''
        q = {}
        device = None
        devices = self._get_quota_devices()
        cmd = [QuotaWrapper.REPQUOTA, '-una']
        data = run_command(cmd)
        grace_regex_pattern = re.compile(
            r'(block|inode)\sgrace\stime:?\s(\d[\w:]+)(?:;|$|\s)',
            re.IGNORECASE)
        for line in data.splitlines():
            if line.startswith('#'):
                if not device:
                    continue
                parts = line.split()
                if len(parts) != 8:
                    parts = self._remove_redundant_fields_from_input(parts)
                uid = parts[0][1:]
                if uid == '0': # We do not want to limit root :)
                    continue
                try:
                    if device not in devices:
                        device = self._find_unknown_device(device)
                    if device in devices and self._is_home_device(
                            self._fetch_homedir(uid), device):
                        q[uid] = dict(map(
                            (lambda x: (self._fields[x[0]],x[1])),
                            enumerate(parts[2:])))
                except (KeyError, NoSuchUserException):
                    continue
            elif line.startswith('***'):
                device = line[line.find('/dev'):].strip()
            elif 'grace' in line:
                found = grace_regex_pattern.findall(line)
                if found:
                    self._grace.update(
                        dict(map((lambda x:(x[0].lower(), x[1])), found)))
        q.update(self._add_default())
        return q

    def _remove_redundant_fields_from_input(self, parts):
        stripped_parts = parts[:2]
        is_digit_pattern=re.compile(r'^\d+$')
        stripped_parts.extend(
            [field for field in parts[2:] if is_digit_pattern.search(field)])
        return stripped_parts

    def _fetch_homedir(self, uid):
        if len(self._uid_to_homedir_map) == 0:
            self._uid_to_homedir_map.update(
                dict(
                    ((str(entry.pw_uid), entry.pw_dir)
                        for entry in pwd.getpwall())))
        try:
            return self._uid_to_homedir_map[uid]
        except KeyError:
            raise NoSuchUserException(uid)

    def _load_quota_devices(self):
        '''
        Gets mounted filesystems list and picks ones with quota on
        '''
        devices = {}
        cmd = [QuotaWrapper.MOUNT]
        patt = re.compile(r'''
            (.*)\son                # filesystem
            \s(\S+)                 # mountpoint
            [^(]*\(                 # find open paren
            .*?usrj?quota
            (?:
                =([\w\.]+)          # get quota file if any
                (?:
                    .*?=(vfs\w+)    # get quota type if any
                )?
            )?
        ''', re.VERBOSE)
        for line in run_command(cmd).splitlines():
            m = patt.match(line)
            if not m:
                continue
            devices[m.group(1)] = {
                'mountpoint':  m.group(2),
                'quota_file':  m.group(3),
                'quota_type':  m.group(4)
            }
        if len(devices) == 0:
            raise QuotaDisabledException('all users', 'all filesystems')
        return devices

    def _load_user_cache(self):
        '''
        For non-privileged user we outputting data from the file
        '''
        q = {}
        try:
            fo = open(QuotaWrapper.CACHEFILE)
            cvs_in = csv.reader(fo, delimiter=',')
        except (OSError, IOError):
            # We don't want to confuse a panel with error messages.
            # Let the data be zeroes until they arrive
            return {str(os.geteuid()):dict.fromkeys(self._fields,'0')}
        uid = str(os.geteuid())
        for row in cvs_in:
            if row[0] == uid:
                q.update({row[0]: dict(map(
                    (lambda x:(self._fields[x], row[x+1])),
                        xrange(len(self._fields))))})
                break
        # We want to prevent crazy cases like misedited cache file
        if not q:
            return {str(os.geteuid()):dict.fromkeys(self._fields,'0')}
        return q

    def _get_home_device(self, home):
        '''
        Returns device user homedir is on
        '''
        device = None
        devices = self._get_quota_devices()
        device_mountpoint_pairs = sorted(
            ((x, devices[x]['mountpoint']) for x in devices.keys()),
            reverse=True, key=operator.itemgetter(1))
        for pair in device_mountpoint_pairs:
            if os.path.dirname(home).startswith(pair[1]):
                device = pair[0]
                break
        return device

    def _is_home_device(self, home, device):
        '''
        Checks if a device is user homedir device
        '''
        return self._get_home_device(home) == device

    def _find_unknown_device(self, device):
        devices = self._get_quota_devices()
        try:
            dev = os.stat(device)[ST_DEV]
            dev_to_find = (os.major(dev), os.minor(dev))
            for current_device in devices.keys():
                dev = os.stat(current_device)[ST_DEV]
                if dev_to_find == (os.major(dev), os.minor(dev)):
                    return current_device
        except OSError:
            return device

    def _add_default(self):
        '''
        Insert 'default' quota
        '''
        values = ['-', '0', '0', '-']
        try:
            values.extend(self._dh.get('users','0').split(':'))
        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
            values.extend(['0', '0'])
        return {'0': dict(map((lambda x: (x[1], values[x[0]])),
            enumerate(self._fields)))}

    def _save_user_limits(self, uid, soft, hard):
        '''
        Saves user limits
        '''
        if soft is None:
            soft = '0'
        if hard is None:
            hard = '0'
        soft, hard = self._get_saved_user_limits_if_none(uid, soft, hard)
        if (soft is None or soft == '0') and (hard is None or hard == '0'
                and self._dh.has_section('users')):
            self._dh.remove_option('users', uid)
        else:
            if not self._dh.has_section('users'):
                self._dh.add_section('users')
            self._dh.set('users', uid, '%s:%s' % (soft, hard))
        self._write_data()

    def _save_package_limits(self, package, soft, hard):
        '''
        Saves package limits
        '''
        if soft is None:
            soft = '0'
        if hard is None:
            hard = '0'
        soft, hard = self._get_saved_package_limits_if_none(package, soft, hard)
        if (soft is None or soft == '0') and (hard is None or hard == '0'
                and self._dh.has_section('packages')):
            self._dh.remove_option('packages', package)
        else:
            if not self._dh.has_section('packages'):
                self._dh.add_section('packages')
            self._dh.set('packages', package, '%s:%s' % (soft, hard))
        self._write_data()

    def _save_data(self, soft, hard, item, item_type):
        '''
        Saves data to a file
        '''
        if soft == '0' and hard == '0':
            try:
                self._dh.remove_option(item_type, item)
            except ConfigParser.NoSectionError:
                pass
        else:
            if not self._dh.has_section(item_type):
                self._dh.add_section(item_type)
            self._dh.set(item_type, item, '%s:%s' % (soft, hard))
        self._write_data()

    def _prepare_writer(self, filepath):
        '''
        Open temporaty file for writing and return file object
        '''
        path = os.path.dirname(filepath)
        try:
            fd, temp_path = tempfile.mkstemp(prefix='lvetmp_', dir=path)
            file_handler = os.fdopen(fd, 'wb')
            self._tmp = temp_path
            return file_handler
        except (IOError, OSError):
            if os.path.exists(temp_path):
                os.unlink(temp_path)
            raise GeneralException("Could not save data")

    def _end_writer(self, path):
        '''
        Routines after writing to file
        '''
        try:
            mask = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH
            os.rename(self._tmp, path)
            os.chmod(path, mask)
        except OSError:
            pass

    def _write_data(self):
        '''
        Actual place of saving data to a file
        '''
        file_handler = self._prepare_writer(QuotaWrapper.DATAFILE)
        self._dh.write(file_handler)
        self._end_writer(QuotaWrapper.DATAFILE)