#!/usr/bin/python

import os
import unittest
import ConfigParser
import random
import pwd
import operator

from clquota import QuotaWrapper
from clquota.utils import run_command    

class TestQuotaWrapper(QuotaWrapper):
    def __init__(self, datafile='/tmp/cltestdata.dat'):
        QuotaWrapper.__init__(self, datafile=datafile)

    def _get_package_to_users_map(self):
        '''
        Replace real method with a stub
        '''
        return {}
    
    def _get_mockup_package(self, mockuser):
        '''
        Instead of real packages mockup is being put
        '''
        mockpackage = 'clquotatestpackage'
        self._user_to_packages_map[mockuser] = []
        self._user_to_packages_map[mockuser].append(mockpackage)
        self._package_to_users_map[mockpackage] = []
        self._package_to_users_map[mockpackage].append(mockuser)
        return mockpackage
    

class QuotaWrapperTest(unittest.TestCase):

    def test_do_user_initial_limits_match(self):
        '''
        Simply compare limits received directly and via object tested
        '''
        user = create_test_user()
        upw = pwd.getpwnam(user)
        uid, home = str(upw.pw_uid), upw.pw_dir
        cl = get_current_limits(uid, home)
        tqw = TestQuotaWrapper(datafile='/tmp/cltestdata.dat')
        tl = tqw.get_limits(uid)
        self.assertTrue(is_dicts_equal(cl, tl[uid]),
            'Initial object limits do not match')
    
    
    def test_have_direct_user_limits_applied(self):
        '''
        Simply apply user limits and check what we got (and in datafile too)
        '''
        datafile = '/tmp/cltestdata.dat'
        clear_saved_data(datafile,'direct')
        soft, hard = '1973', '1975'
        user = create_test_user()
        upw = pwd.getpwnam(user)
        uid, home = str(upw.pw_uid), upw.pw_dir
        tqw = TestQuotaWrapper(datafile)
        tqw._get_mockup_package(uid)
        tqw.set_user_limit(uid, soft, hard)
        cur_limits = get_current_limits(uid, home)
        user_saved = get_saved_info(datafile, 'users', uid)
        self.assertTrue(soft == cur_limits['inodes_soft']
                and hard == cur_limits['inodes_hard']
                and user_saved == '%s:%s' % (soft,hard),
            'Failed to apply direct user quotas (%s)' % (uid,))
        
        
    def test_have_inherited_user_limits_applied(self):
        datafile = '/tmp/cltestdata.dat'
        clear_saved_data(datafile,'inherited')
        user_soft, user_hard = '0', '0'
        pack_soft, pack_hard = '2005', '2011'
        user = create_test_user()
        upw = pwd.getpwnam(user)
        uid, home = str(upw.pw_uid), upw.pw_dir
        tqw = TestQuotaWrapper(datafile)
        mockpack = tqw._get_mockup_package(uid)
        tqw.set_user_limit(uid, user_soft, user_hard)
        cur_limits_before = get_current_limits(uid, home)
        user_saved = get_saved_info(datafile, 'users', uid)
        tqw = None
        tqw = TestQuotaWrapper(datafile)
        mockpack = tqw._get_mockup_package(uid)
        tqw.set_package_limit(mockpack, pack_soft, pack_hard)
        cur_limits_after = get_current_limits(uid, home)
        pack_saved = get_saved_info(datafile, 'packages', mockpack)
        self.assertTrue(user_soft == cur_limits_before['inodes_soft']
                and user_hard == cur_limits_before['inodes_hard']
                and user_saved == '-:-'
                and pack_soft == cur_limits_after['inodes_soft']
                and pack_hard == cur_limits_after['inodes_hard']
                and pack_saved == '%s:%s' % (pack_soft,pack_hard),
            'Failed to apply inherited user quotas (%s)' % (uid,))


    def test_have_default_user_limits_applied(self):
        datafile = '/tmp/cltestdata.dat'
        clear_saved_data(datafile)
        user_soft, user_hard = '0', '0'
        dflt_soft, dflt_hard = '257', '258'
        user = create_test_user()
        upw = pwd.getpwnam(user)
        uid, home = str(upw.pw_uid), upw.pw_dir
        tqw = TestQuotaWrapper(datafile)
        mockpack = tqw._get_mockup_package(uid)
        tqw.set_user_limit(uid, user_soft, user_hard)
        cur_limits_before = get_current_limits(uid, home)
        user_saved = get_saved_info(datafile, 'users', uid)
        tqw = None
        tqw = TestQuotaWrapper(datafile)
        mockpack = tqw._get_mockup_package(uid)
        tqw.set_user_limit('0', dflt_soft, dflt_hard)
        cur_limits_after = get_current_limits(uid, home)
        dflt_saved = get_saved_info(datafile, 'users', '0')
        self.assertTrue(user_soft == cur_limits_before['inodes_soft']
                and user_hard == cur_limits_before['inodes_hard']
                and user_saved == '-:-'
                and dflt_soft == cur_limits_after['inodes_soft']
                and dflt_hard == cur_limits_after['inodes_hard']
                and dflt_saved == '%s:%s' % (dflt_soft,dflt_hard),
            'Failed to apply default user quotas (%s)' % (uid,))

    #def test_have_package_limits_applied_when_no_user_limits(self):
    #    datafile = '/tmp/cltestdata.dat'
    #    clear_saved_data(datafile,'inherited')
    #    user_soft, user_hard = '1290', '1291'
    #    pack_soft, pack_hard = '3005', '3011'
    #    user = create_test_user()
    #    upw = pwd.getpwnam(user)
    #    uid, home = str(upw.pw_uid), upw.pw_dir
    #    tqw = TestQuotaWrapper(datafile)
    #    mockpack = tqw._get_mockup_package(uid)
    #    save_data(datafile, 'packages', mockpack, '%s:%s' % (pack_soft, pack_hard))
    #    with open(datafile) as f:
    #        print f.read()
    #    tqw.set_user_limit(uid, user_soft, user_hard)
    #    with open(datafile) as f:
    #        print f.read()
    #    cur_limits_before = get_current_limits(uid, home)
    #    user_saved = get_saved_info(datafile, 'users', uid)
    #    tqw = None
    #    tqw = TestQuotaWrapper(datafile)
    #    mockpack = tqw._get_mockup_package(uid)
    #    tqw.set_package_limit(mockpack, pack_soft, pack_hard)
    #    cur_limits_after = get_current_limits(uid, home)
    #    pack_saved = get_saved_info(datafile, 'packages', mockpack)
    #    self.assertTrue(user_soft == cur_limits_before['inodes_soft']
    #            and user_hard == cur_limits_before['inodes_hard']
    #            and user_saved == '-:-'
    #            and pack_soft == cur_limits_after['inodes_soft']
    #            and pack_hard == cur_limits_after['inodes_hard']
    #            and pack_saved == '%s:%s' % (pack_soft,pack_hard),
    #        'Failed to apply inherited user quotas (%s)' % (uid,))

def create_test_user():
    try:
        return create_test_user.user
    except AttributeError:
        pass
    name = 'clqt' + str(random.random())[2:]
    run_command(['/usr/sbin/useradd', '-s', '/bin/false', name])
    create_test_user.user = name
    return name

def delete_test_user(name):
    home = get_user_attr(name)
    run_command(['/usr/sbin/userdel', name])
    delattr(create_test_user, 'user')
    run_command(['rm', '-rf', home])

def get_devices():
    data = {}
    mounts = run_command(['/bin/mount'])
    for line in mounts.splitlines():
        if not line.startswith('/dev'):
            continue
        pair = map((lambda x: x.strip()),line[:line.find('type')].split('on'))
        data[pair[0]] = pair[1]
    return data

def get_home_device(home):
        device = None
        devices = get_devices()
        devices = sorted(((x, devices[x]) for x in devices.keys()),
            reverse=True, key=operator.itemgetter(1))
        for pair in devices:
            if os.path.dirname(home).startswith(pair[1]):
                device = pair[0]
                break
        return device

def get_current_limits(uid, home):
    device = get_home_device(home)
    limits = run_command(['/usr/sbin/repquota', '-un', device])
    for line in limits.splitlines():
        if not line.startswith('#'+uid):
            continue
        return parse_quota_line(line)

def set_user_limits(uid, soft, hard, home):
    device = get_home_device(home)
    l = get_current_limits(uid, home)
    run_command(['/usr/sbin/setquota', '-u', l['bytes_soft'], l['bytes_hard'],
        soft, hard, device])

def parse_quota_line(line):
    f = ['bytes_used','bytes_soft','bytes_hard',
        'inodes_used','inodes_soft','inodes_hard']
    return dict(map((lambda x: (f[x[0]],x[1])), enumerate(line.split()[2:])))

def is_dicts_equal(d1, d2):
    if len(d1.keys()) != len(d2.keys()):
        return False
    for key in d1.keys():
        try:
            if d1[key] != d2[key]:
                return False
        except KeyError:
            return False
    return True

def get_saved_info(path, item_type, item):
    cp = ConfigParser.ConfigParser()
    cp.read(path)
    if cp.has_section(item_type) and cp.has_option(item_type, item):
        return cp.get(item_type, item)
    return '-:-'
    
def clear_saved_data(path,msg=None):
    try:
        os.unlink(path)
    except (OSError, IOError):
        if msg:
            print "CANNOT DELETE FILE FROM %s" % (msg,)
        pass

def save_data(path, item_type, item, data):
    cp = ConfigParser.ConfigParser()
    cp.read(path)
    if not cp.has_section(item_type):
        cp.add_section(item_type)
    cp.set(item_type, item, data)
    fd = open(path, 'wb')
    cp.write(fd)
    fd.close()


if __name__ == '__main__':
    unittest.main()